/*
  cadaverserver
  Copyright (C) 1999  Erik Thiele

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "cyborg.hpp"

#include <stdio.h>

cyborg::cyborg()
{
  for (int i=0;i<boolvalue_count();i++) boolvalue[i]=false;
  for (int i=0;i<intvalue_count();i++) intvalue[i]=0;
  for (int i=0;i<floatvalue_count();i++) floatvalue[i]=0;
};

const char *const cyborg::boolvalue_textdescription (const boolvalue_type number)
{
  switch (number) {
  case ISMALE: return "ismale";
  default: return "unknown";
  };
};

const char *const cyborg::intvalue_textdescription (const intvalue_type number)
{
  switch (number) {
  case STATE: return "state";
  case TAKEDROP: return "takedrop";
  case TARGETID: return "targetid";
  case ID: return "id";
  case OWNERID: return "ownerid";
  default: return "unknown";
  };
};

const char *const cyborg::floatvalue_textdescription (const floatvalue_type number)
{
  switch (number) {
  case FERTILITY: return "fertility";
  case HEALTH: return "health";
  case AGE: return "age";
  case FIGHTSTRENGTH: return "fightstrength";
  case TARGETX: return "targetx";
  case TARGETY: return "targety";
  case POSX: return "posx";
  case POSY: return "posy";
  case WATER: return "water";
  case GRAIN: return "grain";
  case GRASS: return "grass";
  case CARCASS: return "carcass";
  case HAY: return "hay";
  case MEAT: return "meat";
  case BREAD: return "bread";
  case FOREST: return "forest";
  default: return "unknown";
  };
};

const char *const cyborg::state_textdescription (const state_type number)
{
  switch (number) {
  case IDLE: return "idle";
  case GOTO: return "goto";
  case FIGHT: return "fight";
  case TAKE: return "take";
  case DROP: return "drop";
  case EAT: return "eat";
  case CONVERT: return "convert";
  case BURN: return "burn";
  default: return "unknown";
  };
};

const char *const cyborg::takedrop_textdescription (const takedrop_type number)
{
  switch (number) {
  case TWATER: return "water";
  case TGRAIN: return "grain";
  case TGRASS: return "grass";
  case TCARCASS: return "carcass";
  case THAY: return "hay";
  case TMEAT: return "meat";
  case TBREAD: return "bread";
  case TFOREST: return "forest";
  default: return "unknown";
  };
};

void cyborg::dump (char *buf) const
{
  int pos=0, i;
  for (i=0; i<boolvalue_count(); i++)
    pos+=sprintf(buf+pos, "%c ", boolvalue[i] ? 'T' : 'F');
  for (i=0; i<intvalue_count(); i++)
    pos+=sprintf(buf+pos, "%d ",intvalue[i]);
  for (i=0; i<floatvalue_count(); i++)
    pos+=sprintf(buf+pos, "%.1f ",floatvalue[i]);
  pos--;
  buf[pos] = 0;
};

void cyborg::load (const char *buf)
{
  int pos=0, i;
  for (i=0; i<boolvalue_count(); i++) {
    boolvalue[i] = buf[pos] == 'T' ? true : false;
    pos+=2;
  };
  for (i=0; i<intvalue_count(); i++) {
    if (sscanf(buf+pos,"%d",&intvalue[i])!=1);
    while (buf[pos++]>32);
  };
  for (i=0; i<floatvalue_count(); i++) {
    if (sscanf(buf+pos,"%f",&floatvalue[i])!=1);
    while (buf[pos++]>32);
  };
};


spielcyborg::spielcyborg ()
{
};

spielcyborg::~spielcyborg ()
{
  clear();
};

void spielcyborg::add (cyborg *const newcyborg)
{
  all.insert (pair<const int, cyborg *const>(newcyborg->get_int(cyborg::ID),newcyborg));
};

void spielcyborg::del (const int id)
{
  delete all[id];
  all.erase (id);
};

void spielcyborg::clear()
{
  for (map<const int, cyborg *const>::iterator i=all.begin();i!=all.end();i++)
    delete i->second;
  all.clear();
};
