/*
  catrats
  Copyright (C) 1999  Erik Thiele

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "spielfeld.hpp"
#include "cyborg.hpp"
#include "yyyzeit.hpp"
#include "memchk.hpp"
#include "selector.hpp"

#include <math.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include <fcntl.h>
#include <time.h>
#include <stdlib.h>
#include <string>
#include <vector>
#include <algorithm>

using namespace yyyzeit;

int comm,myid;
char clip[10000];
spielfeld *gsf=0;
spielcyborg *gsc=0;
string *outputbuffer_ptr=0;
#define outputbuffer (*outputbuffer_ptr)

float hqx=0.5, hqy=0.5;
bool armaggedon=false;
bool havepeople=false;
bool ende=false;
bool debug=false;

struct pfeldtype {
  int x,y,prio;
};


struct pfeld_farmersort : public binary_function<const pfeldtype&,const pfeldtype&,const bool>
{
  const bool operator() (const pfeldtype &a, const pfeldtype &b)
  {
    float adx = a.x - hqx;
    float ady = a.y - hqy;
    float bdx = b.x - hqx;
    float bdy = b.y - hqy;
    float da = adx*adx + ady*ady;
    float db = bdx*bdx + bdy*bdy;
    return da < db;
  }
};

struct pfeld_thieffoodfetchersort : public binary_function<const pfeldtype&,const pfeldtype&,const bool>
{
  pfeld_thieffoodfetchersort(cyborg *c) 
  {
    cx=c->get_float(cyborg::POSX);
    cy=c->get_float(cyborg::POSY);
  };
  const bool operator() (const pfeldtype &a, const pfeldtype &b)
  {
    float adx = a.x - cx;
    float ady = a.y - cy;
    float bdx = b.x - cx;
    float bdy = b.y - cy;
    float da = adx*adx + ady*ady;
    float db = bdx*bdx + bdy*bdy;
    return da < db;
  }
  float cx, cy;
};

void usage()
{
  fprintf(stderr,"Usage: catrats serveradress\n\n");
  exit(1);
};

void errabo(const char *const msg)
{
  fprintf(stderr,"\n\nerror.\n%s\n\ntrying to dump core\n\n",msg);
  raise (SIGSEGV);
};

void erraboerrno(const char *const msg)
{
  fprintf(stderr,"\n\nerror.\n%s\nerrno is %s\n\ntrying to dump core\n\n",
	  msg,strerror(errno));
  raise (SIGSEGV);
};


class localtraverser {
public:
  localtraverser (int mx_, int my_, int rad_)
    : rad(rad_), mx(mx_), my(my_)
  {
    x=mx-rad - 1; // one before legal.
    y=my-rad;
  };

  bool getpos()
  {
    do {
      if (!movenext()) return false;
    }while (!valid());
    return true;
  };

  int x, y, rad;
  
private:
  bool movenext()
  {
    if (x>mx+rad) return false;
    x++;
    if (x>mx+rad) {
      y++;
      if (y>my+rad) return false;
      x=mx-rad;
    };
    return true;
  };

  bool valid()
  {
    return ((x>=0) && (x<gsf->xsize()) && (y>=0) && (y<gsf->ysize()));
  };

  int mx, my;
};

float distance(cyborg *a, cyborg *b)
{
  return hypot(a->get_float(cyborg::POSX) - b->get_float(cyborg::POSX),
	       a->get_float(cyborg::POSY) - b->get_float(cyborg::POSY) );
};

bool isfieldaligned(cyborg *c)
{
  float a = fmodf(c->get_float(cyborg::POSX), 1.0),
    b = fmodf(c->get_float(cyborg::POSY), 1.0);
  return (a>0.25) && (a<0.75) && (b>0.25) && (b<0.75);
};

void getclosestfieldcoords(cyborg *c, int &x, int &y)
{
  x=int(c->get_float(cyborg::POSX));
  y=int(c->get_float(cyborg::POSY));
};

void getfieldcoords(cyborg *c, int &x, int &y)
{
  if (!isfieldaligned(c)) errabo("getfieldcoords(...): isfieldaligned == false");
  getclosestfieldcoords(c,x,y);
};

bool start_eat(cyborg *c)
{
  if (c->get_float(cyborg::BREAD)>0.1) {
    sprintf(clip,"EA %d %d\n",c->get_int(cyborg::ID),cyborg::TBREAD); outputbuffer+=clip;
    return true;
  }else if (c->get_float(cyborg::MEAT)>0.1) {
    sprintf(clip,"EA %d %d\n",c->get_int(cyborg::ID),cyborg::TMEAT); outputbuffer+=clip;
    return true;
  };
  return false;
};

bool emergency_eat_condition(cyborg *c)
{
  return c->get_float(cyborg::HEALTH)<0.7;
};

bool emergency_eat(cyborg *c)
{
  if (emergency_eat_condition(c)) {
    if (start_eat(c)) {
      //      printf("emergency_eat: started to eat, id=%d\n",c->get_int(cyborg::ID));
      return true;
    }else{
      return false;
    };
  };
  return false;
};

bool emergency_runhome(cyborg *c)
{
  if (c->get_float(cyborg::HEALTH)<0.5) {
    if (hypot(hqx - c->get_float(cyborg::POSX), hqy - c->get_float(cyborg::POSY)) > 7.0) {
      sprintf(clip,"MG %d %f %f\n",c->get_int(cyborg::ID),hqx,hqy); outputbuffer+=clip;
      //      printf("emergency_runhome: running home, id=%d\n",c->get_int(cyborg::ID));
      return true;
    } else {
      return false;
    };
  } else {
    return false;
  };
};

bool normal_eat(cyborg *c)
{
  if (c->get_float(cyborg::HEALTH)<1.2) return start_eat(c);
  return false;
};

bool stoptogoto_foreat (cyborg *c)
{
  if (emergency_eat_condition(c)) {
    float shortest=1e100;
    cyborg *shortestcyb=0;
    for (map<const int,cyborg *const>::iterator sc=gsc->all.begin();sc!=gsc->all.end();sc++) {
      if (sc->second->get_int(cyborg::OWNERID)!=myid) {
	if (distance(c,sc->second) < shortest) {
	  shortest = distance(c,sc->second);
	  shortestcyb = sc->second;
	};
      };
    };
    
    if ((shortestcyb)&&(shortest<3.0)) {
      return false;
    } else {
      if (start_eat(c)) {
	//	printf("stoptogoto_foreat: aborting goto for emergency eat, id=%d\n",c->get_int(cyborg::ID));
	return true;
      } else {
	return false;
      };
    };
  } else {
    return false;
  };
};

bool stoptoeat_forgoto (cyborg *c)
{
  if (emergency_eat_condition(c)) {
    float shortest=1e100;
    cyborg *shortestcyb=0;
    for (map<const int,cyborg *const>::iterator sc=gsc->all.begin();sc!=gsc->all.end();sc++) {
      if (sc->second->get_int(cyborg::OWNERID)!=myid) {
	if (distance(c, sc->second) < shortest) {
	  shortest = distance(c, sc->second);
	  shortestcyb = sc->second;
	};
      };
    };
    
    if ((shortestcyb)&&(shortest<2.0)) {
      if (emergency_runhome(c)) {
	//	printf("stoptoeat_forgoto: aborting eat for emergency runhome, id=%d\n",c->get_int(cyborg::ID));
	return true;
      } else {
	return false;
      };
    } else {
      return false;
    };
  } else {
    return false;
  };
};

bool burner(cyborg *c)
{
  if (c->get_float(cyborg::WATER)>3.0) {
    sprintf(clip,"BU %d %d\n",c->get_int(cyborg::ID),cyborg::TWATER); outputbuffer+=clip; return true;
  };
  if (c->get_float(cyborg::GRASS)>3.0) {
    sprintf(clip,"BU %d %d\n",c->get_int(cyborg::ID),cyborg::TGRASS); outputbuffer+=clip; return true;
  };
  if (c->get_float(cyborg::HAY)>3.0) {
    sprintf(clip,"BU %d %d\n",c->get_int(cyborg::ID),cyborg::THAY); outputbuffer+=clip; return true;
  };
  if (c->get_float(cyborg::FOREST)>3.0) {
    sprintf(clip,"BU %d %d\n",c->get_int(cyborg::ID),cyborg::TFOREST); outputbuffer+=clip; return true;
  };
  return false;
};

bool dropper(cyborg *c)
{
  if (isfieldaligned(c)) {
    int x,y;
    getfieldcoords(c,x,y);
    if ((c->get_float(cyborg::BREAD)>2.0)&&(gsf->get(x,y).get_float(feld::BREAD)<0.9)) {
      sprintf(clip,"DR %d %d\n",c->get_int(cyborg::ID),cyborg::TBREAD); outputbuffer+=clip; return true;
    };
    if ((c->get_float(cyborg::MEAT)>2.0)&&(gsf->get(x,y).get_float(feld::MEAT)<0.9)) {
      sprintf(clip,"DR %d %d\n",c->get_int(cyborg::ID),cyborg::TMEAT); outputbuffer+=clip; return true;
    };
  };
  return false;
};

bool fighting(cyborg *c)
{
  float shortest=1e100;
  cyborg *shortestcyb=0;
  float myx=c->get_float(cyborg::POSX);
  float myy=c->get_float(cyborg::POSY);

  for (map<const int,cyborg *const>::iterator sc=gsc->all.begin();sc!=gsc->all.end();sc++) {
    if (sc->second->get_int(cyborg::OWNERID)!=myid) {
      float itx=sc->second->get_float(cyborg::POSX);
      float ity=sc->second->get_float(cyborg::POSY);
      float mydist = hypotf(myx-itx,myy-ity);
      if (mydist < shortest) {
	shortest=mydist;
	shortestcyb=sc->second;
      };
    };
  };

  if (shortestcyb) {
    if (shortest<1.5) {
      sprintf(clip,"MF %d %d\n",c->get_int(cyborg::ID),shortestcyb->get_int(cyborg::ID)); outputbuffer+=clip;
      return true;
    }else if ((shortest<5.0) || (armaggedon)) {
      float itx=shortestcyb->get_float(cyborg::POSX);
      float ity=shortestcyb->get_float(cyborg::POSY);
      sprintf(clip,"MG %d %f %f\n",c->get_int(cyborg::ID),(myx+itx)/2.0,(myy+ity)/2.0); outputbuffer+=clip;
      return true;
    };
  };

  return false;
};

bool fertility(cyborg *c)
{
  if (!c->get_bool(cyborg::ISMALE)) return false;
  if (c->get_float(cyborg::FERTILITY)>0.1) return false;

  float shortest=1e100;
  cyborg *shortestcyb=0;
  float myx=c->get_float(cyborg::POSX);
  float myy=c->get_float(cyborg::POSY);

  for (map<const int,cyborg *const>::iterator sc=gsc->all.begin();sc!=gsc->all.end();sc++) {
    if ((sc->second->get_int(cyborg::OWNERID)==myid)
	&& (!sc->second->get_bool(cyborg::ISMALE))
	&& (sc->second->get_float(cyborg::FERTILITY)<0.1))
      {
      float itx=sc->second->get_float(cyborg::POSX);
      float ity=sc->second->get_float(cyborg::POSY);
      float mydist = hypotf(myx-itx,myy-ity);
      if (mydist < shortest) {
	shortest=mydist;
	shortestcyb=sc->second;
      };
    };
  };

  if (shortestcyb) {
    if (shortest<1.5) {
      sprintf(clip,"MS %d %d\n",c->get_int(cyborg::ID),shortestcyb->get_int(cyborg::ID)); outputbuffer+=clip;
      return true;
    }else if (shortest<7.0) {
      float itx=shortestcyb->get_float(cyborg::POSX);
      float ity=shortestcyb->get_float(cyborg::POSY);
      sprintf(clip,"MG %d %f %f\n",c->get_int(cyborg::ID),(myx+itx)/2.0,(myy+ity)/2.0); outputbuffer+=clip;
      return true;
    };
  };

  return false;
};

bool converter(cyborg *c)
{
  if (c->get_float(cyborg::GRAIN)>0.1) {
    sprintf(clip,"CO %d %d\n",c->get_int(cyborg::ID),cyborg::TBREAD); outputbuffer+=clip;
    return true;
  };
  if (c->get_float(cyborg::CARCASS)>0.1) {
    sprintf(clip,"CO %d %d\n",c->get_int(cyborg::ID),cyborg::TMEAT); outputbuffer+=clip;
    return true;
  };
  return false;
};

bool thief_foodfetcher_condition(cyborg *c)
{
  return (c->get_float(cyborg::BREAD) + c->get_float(cyborg::MEAT)*2.0) < 0.5;
};

bool thief_foodfetcher_goodfield(int x, int y)
{
  return (gsf->get(x,y).get_float(feld::BREAD)>0.1)
    || (gsf->get(x,y).get_float(feld::MEAT)>0.1);
};

void thief_foodfetcher_dojob(int id, int x, int y)
{
  if (gsf->get(x,y).get_float(feld::BREAD)>0.1) {
    sprintf(clip,"TA %d %d\n",id,cyborg::TBREAD); outputbuffer+=clip;
  }else if (gsf->get(x,y).get_float(feld::MEAT)>0.1) {
    sprintf(clip,"TA %d %d\n",id,cyborg::TMEAT); outputbuffer+=clip;
  }else errabo("thief_foodfetcher_dojob(...): nothing to do");
};

bool thief_foodfetcher(cyborg *c)
{
  if (thief_foodfetcher_condition(c)) {
    int x,y;
    if (isfieldaligned(c)) {
      getfieldcoords(c,x,y);
      if (thief_foodfetcher_goodfield(x,y)) {
	thief_foodfetcher_dojob(c->get_int(cyborg::ID),x,y);
	return true;
      };
    }else{
      getclosestfieldcoords(c,x,y);
    };
    localtraverser t(x,y,9);
    vector<pfeldtype> pfeld;
    while (t.getpos()) {
      pfeldtype a;
      a.x=t.x;
      a.y=t.y;
      a.prio=thief_foodfetcher_goodfield(t.x,t.y) ? 1 : 0;
      if (a.prio!=0) pfeld.push_back(a);
    };
    if (pfeld.size()) {
      stable_sort(pfeld.begin(),pfeld.end(),pfeld_thieffoodfetchersort(c));
      int divi=pfeld.size()/2;
      if (divi<1) divi=1;
      int i=rand() % divi;
      sprintf(clip,"MG %d %f %f\n",c->get_int(cyborg::ID),pfeld[i].x+0.5,pfeld[i].y+0.5); outputbuffer+=clip;
      return true;
    };
    return false;
  };
  return false;
};

int farming_priority (int x, int y)
{
  // hier steckt noch intelligenzmoeglichkeit FIXME
  if (gsf->get(x,y).get_float(feld::CARCASS)>0.1) return 4;
  if (gsf->get(x,y).get_float(feld::GRAIN)>0.9) return 3;
  if (gsf->get(x,y).get_float(feld::GRASS)>0.1) return 2;
  if (gsf->get(x,y).get_float(feld::FOREST)>0.1) return 1;
  return 0;
};

void do_farming(int id, int x, int y)
{
  int prio=farming_priority(x,y);
  if (prio==0) errabo("do_farming: nothing to do");
  if (prio==1) {
    sprintf(clip,"TA %d %d\n",id,cyborg::TFOREST); outputbuffer+=clip;
  }else if (prio==2) {
    sprintf(clip,"TA %d %d\n",id,cyborg::TGRASS); outputbuffer+=clip;
  }else if (prio==3) {
    sprintf(clip,"TA %d %d\n",id,cyborg::TGRAIN); outputbuffer+=clip;
  }else if (prio==4) {
    sprintf(clip,"TA %d %d\n",id,cyborg::TCARCASS); outputbuffer+=clip;
  }else errabo("do_farming: unknown priority");
};

bool farmer (cyborg *c)
{
  int x,y;
  if (isfieldaligned(c)) {
    getfieldcoords(c,x,y);
    if (farming_priority(x,y)) {
      do_farming(c->get_int(cyborg::ID), x, y);
      return true;
    };
  }else{
    getclosestfieldcoords(c,x,y);
  };
  localtraverser t(x,y,2);
  vector<pfeldtype> pfeld;
  while (t.getpos()) {
    pfeldtype a;
    a.x=t.x;
    a.y=t.y;
    a.prio=farming_priority(t.x,t.y);
    if (a.prio!=0) pfeld.push_back(a);
  };
  if (pfeld.size()) {
    stable_sort(pfeld.begin(),pfeld.end(),pfeld_farmersort());
    int divi=pfeld.size()/2;
    if (divi<1) divi=1;
    int i=rand() % divi;
    sprintf(clip,"MG %d %f %f\n",c->get_int(cyborg::ID),pfeld[i].x+0.5,pfeld[i].y+0.5); outputbuffer+=clip;
    return true;
  };
  return false;
};

bool harvested_enough_grain(cyborg *c)
{
  int x,y;
  getfieldcoords(c,x,y);
  if (gsf->get(x,y).get_float(feld::GRAIN)<0.6) return true;
  return false;
};

bool have_enough_meat(cyborg *c)
{
  return c->get_float(cyborg::MEAT)>1.0;
};

bool have_enough_bread(cyborg *c)
{
  return c->get_float(cyborg::BREAD)>1.0;
};

void idle(cyborg *c)
{
  sprintf(clip,"MI %d\n",c->get_int(cyborg::ID)); outputbuffer+=clip;
};

void sethq(const float newhqx, const float newhqy)
{
  hqx=newhqx;
  hqy=newhqy;
  printf("i set the headquarter position to %f %f\n",hqx,hqy);
};

void think ()
{
  if (havepeople) {
    if (gsc->all.empty()) {
      printf("\n\nI have no more people. I give up.\n\n");
      ende=true;
    };
  }else{
    if (!gsc->all.empty()) {
      havepeople=true;
      sethq( gsc->all.begin()->second->get_float(cyborg::POSX),
	     gsc->all.begin()->second->get_float(cyborg::POSY) );
    };
  };

  for (map<const int,cyborg *const>::iterator it=gsc->all.begin();it!=gsc->all.end();it++) {
    cyborg *c = it->second;
    if (c->get_int(cyborg::OWNERID)==myid) {
      switch (c->get_int(cyborg::STATE)) {
      case cyborg::IDLE:
	if (emergency_eat(c)) break;
	if (emergency_runhome(c)) break;
	if (normal_eat(c)) break;
	if (thief_foodfetcher(c)) break;
	if (burner(c)) break;
	if (dropper(c)) break;
	
	if (fighting(c)) break;
	if (fertility(c)) break;
	if (converter(c)) break;
	if (farmer(c)) break;
	break;
      case cyborg::GOTO:
	if (stoptogoto_foreat(c)) break;
	break;
      case cyborg::FIGHT:
	if (emergency_eat(c)) break;
	if (emergency_runhome(c)) break;
	break;
      case cyborg::TAKE:
	if (emergency_eat(c)) break;
	if (emergency_runhome(c)) break;
	switch (c->get_int(cyborg::TAKEDROP)) {
	case cyborg::TWATER: errabo ("i am taking water");
	case cyborg::TGRAIN: if (harvested_enough_grain(c)) idle(c); break;
	case cyborg::TGRASS: break;
	case cyborg::TCARCASS: break;
	case cyborg::THAY: errabo ("i am taking hay");
	case cyborg::TMEAT: if (have_enough_meat(c)) idle(c); break;
	case cyborg::TBREAD: if (have_enough_bread(c)) idle(c); break;
	case cyborg::TFOREST: break;
	};
	break;
      case cyborg::DROP:
	if (emergency_eat(c)) break;
	if (emergency_runhome(c)) break;
	if (fighting(c)) break;
	switch (c->get_int(cyborg::TAKEDROP)) {
	case cyborg::TWATER: errabo ("i am dropping water");
	case cyborg::TGRAIN: errabo ("i am dropping grain");
	case cyborg::TGRASS: errabo ("i am dropping grass");
	case cyborg::TCARCASS: errabo ("i am dropping carcass");
	case cyborg::THAY: errabo ("i am dropping hay");
	case cyborg::TMEAT: if (!have_enough_meat(c)) idle(c); break;
	case cyborg::TBREAD: if (!have_enough_bread(c)) idle(c); break;
	case cyborg::TFOREST: errabo ("i am dropping forest");
	};
	break;
      case cyborg::EAT:
	if (stoptoeat_forgoto(c)) break;
	if (c->get_float(cyborg::HEALTH)>=1.7) { idle(c); break; };
	break;
      case cyborg::CONVERT:
	if (emergency_eat(c)) break;
	if (emergency_runhome(c)) break;
	if (fighting(c)) break;
	break;
      case cyborg::BURN:
	if (emergency_eat(c)) break;
	if (emergency_runhome(c)) break;
	if (fighting(c)) break;
	break;
      };
    };
  };
};


void help()
{
  printf(
	 "\n"
	 "\n"
	 "armaggedon   - activate final fight mode.\n"
	 "debug        - toggle data stream debugging\n"
	 "quit         - give up.\n"
	 "help         - give this help\n"
	 "sethq x y    - set headquarter position\n"
	 "\n"
	 );
};

int main(int argc,char *argv[])
{
  memchk::resetsituation(); {

  printf("\nCATRATS ! Artificial Intelligence client ! Version 1.0.1\n\n");
  
  if (argc!=2) usage();

  outputbuffer_ptr=new string;
  signal (SIGPIPE,SIG_IGN);
  fcntl (0,F_SETFL,O_NONBLOCK);
  printf("resolving adress\n");
  struct sockaddr_in name;
  struct hostent *hostinfo;
  name.sin_family = AF_INET;
  name.sin_port = htons (8932);
  hostinfo = gethostbyname (argv[1]);
  if (hostinfo == NULL) errabo("cannot resolve host");
  name.sin_addr = *(struct in_addr *) hostinfo->h_addr;
  if ((comm = socket (PF_INET, SOCK_STREAM, 0))<0) erraboerrno("socket failed");
  printf("connecting\n");
  if (connect (comm,(struct sockaddr *) &name,sizeof (name))<0)
    erraboerrno("connect failed");
  printf("connected !!!!!\n");
  printf("waiting for map size\n");
  int pos=0;
  for (;;pos++) {
    if (read(comm,&clip[pos],1)!=1) erraboerrno("server is stupid");
    if (clip[pos]<32) break;
  };
  clip[pos]=0;
  int nw,nh;
  sscanf(clip,"%d %d %d",&nw,&nh,&myid);
  gsf = new spielfeld(nw,nh);
  printf("width=%d height=%d myid=%d\n",gsf->xsize(),gsf->ysize(),myid);
  fcntl (comm,F_SETFL,O_NONBLOCK);

  srand(time(0));

  gsc=new spielcyborg;

  string inputline, promptline;
  printf("> "); fflush(stdout);

  zeit interval(seconds(1));
  zeit timeout(getcurrenttime());
  bool timeoutactive=true;
  bool flushit=false;

  selector mysel;
  int retval;
  while (!ende) {
    mysel.reset();
    mysel.add_read(0);
    mysel.add_read(comm);
    if (outputbuffer.size()) {
      mysel.add_write(comm);
    }else{
      if (flushit) {
	flushit=false;
	timeout=getcurrenttime()+interval;
	timeoutactive=true;
      };
    };
    if (timeoutactive) mysel.settimeout(timeout);
    mysel.waitnow();

    if (mysel.isready_read(0)) {
      char c;
      int retval;
      while ((retval=read(0,&c,1))==1) {
	if (c<32) {
	  if (promptline=="help") {
	    help();
	  }else if (promptline=="quit") {
	    ende=true;
	  }else if (promptline=="debug") {
	    debug=!debug;
	  }else if (promptline=="armaggedon") {
	    if (armaggedon) {
	      printf("\n\narmaggedon is already activated.\ni told you there is no way back\n\n");
	    }else{
	      armaggedon=true;
	      printf("\n\nARMAGGEDON! There is no way back now.\n\n");
	    };
	  } else if ((promptline.size()>=5) && (promptline.substr(0,5)=="sethq")) {
	    promptline.erase(0,5);
	    float fx,fy;
	    if (sscanf(promptline.c_str(),"%f %f",&fx,&fy)!=2) fprintf(stderr,"wrong parameters\n");
	    else sethq(fx,fy);
	  }else if (promptline.size()) {
	    fprintf(stderr,"unknown command (try 'help')\n");
	  };
	  printf("> "); fflush(stdout);
	  promptline="";
	}else{
	  promptline+=c;
	};
      };
      if (retval==0) errabo("EOF on stdin.");
      else if (errno!=EAGAIN) erraboerrno("read error on stdin");
    };
    
    if (mysel.isready_read(comm)) {
      int retval;
      while ((retval=read(comm,clip,1024))>0) {
	clip[retval]=0;
	inputline+=clip;
      };
      if (retval==0) errabo("serverconnection error");
      else if (errno!=EAGAIN) erraboerrno("serverconnection error");
      
      for (unsigned scan=0;scan<inputline.size();) {
	if (inputline[scan]<32) {
	  if ((scan>=2)&&(inputline.substr(0,2)=="RM")) {
	    int pos=2,nx,ny;
	    for (;;) {
	      while (inputline[pos]==32)pos++;
	      if (inputline[pos]<32) break; // no more changes
	      if (sscanf(inputline.substr(pos,scan-pos).c_str(),"%d %d",&nx,&ny)!=2) errabo("RM msg parse error 1");
	      for (int i=0;i<2;) if (inputline[pos++]<=32) i++;
	      gsf->get(nx,ny).load(inputline.substr(pos,8).c_str());
	      for (int i=0;i<1;) if (inputline[pos++]<=32) i++;
	      pos--;
	    };
	  }else if ((scan>=2)&&(inputline.substr(0,2)=="RC")) {
	    gsc->clear();
	    int pos=2;
	    for (;;) {
	      while (inputline[pos]==32)pos++;
	      if (inputline[pos]<32) break; // no more cyborgs
	      cyborg *p=new cyborg;
	      p->load(inputline.substr(pos,scan-pos).c_str());
	      gsc->add(p);
	      for (int i=0;i<22;) if (inputline[pos++]<=32) i++;
	      pos--;
	    };
	    think();
	    flushit=true;
	  }else if ((scan>=3)&&(inputline.substr(0,3)=="MM ")) {
	    printf("server sent me a text message: %s\n", inputline.substr(3,scan-3).c_str());
	  }else{
	    printf("received unknown command from server:\n%s\n", inputline.substr(0,scan).c_str());
	  };
	  
	  inputline.erase(0,scan+1);
	  scan=0;
	}else{
	  scan++;
	};
      };
    };
  
    if (mysel.isready_write(comm)) {
      while (outputbuffer.size()) {
	retval=write(comm,outputbuffer.c_str(),outputbuffer.size()>1024 ? 1024 : outputbuffer.size());
	if (debug) for (int x=0;x<retval;x++) printf("%c",outputbuffer[x]);
	if (retval==-1) {
	  if (errno==EAGAIN) {
	    break;
	  }else{
	    erraboerrno("output error this sucks.!");
	  };
	}else if (retval==0) {
	  errabo("the server closed my stream ! hehe looks like i am the greatest");
	}else{
	  outputbuffer.erase(0,retval);
	};
      };
    };
    
    if (mysel.istimeoutexpired()) {
      timeoutactive=false;
      outputbuffer+="QN\n";
    };
  };
  
  delete gsc;
  delete gsf;
  delete outputbuffer_ptr;
  close(comm);
 
  } if (!memchk::everythingzero()) {
    cerr << "Memory statistics error !!!" << endl;
    memchk::printstatistics(cerr);
  };

  printf("\nByebye catrats !\n\n");
  return 0;
};


